# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Test end to end flow with:
        2 foglamps, One foglamp use programatic south services, sinusoid, expression and playback to send data
        via http north to send foglamp
        second foglamp use PI Server (C) plugin to send data to PI
"""

import subprocess
import http.client
import os
import json
import time
import pytest
from collections import Counter
import utils


__author__ = "Vaibhav Singhal"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


CSV_NAME = "sample.csv"
CSV_HEADERS = "ivalue"
CSV_DATA = "10,20,21,40"

NORTH_TASK_NAME = "NorthReadingsTo_PI"
remote_asset_name = "fogpair_playback"


class TestE2eFogPairPi:

    def get_asset_list(self, foglamp_url):
        _connection = http.client.HTTPConnection(foglamp_url)
        _connection.request("GET", '/foglamp/asset')
        r = _connection.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        actual_asset_codes = []
        for itm in jdoc:
            actual_asset_codes.append(itm["assetCode"])
        return actual_asset_codes

    def get_ping_status(self, foglamp_url):
        _connection = http.client.HTTPConnection(foglamp_url)
        _connection.request("GET", '/foglamp/ping')
        r = _connection.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        return jdoc

    def get_statistics_map(self, foglamp_url):
        _connection = http.client.HTTPConnection(foglamp_url)
        _connection.request("GET", '/foglamp/statistics')
        r = _connection.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        return utils.serialize_stats_map(jdoc)

    @pytest.fixture
    def reset_and_start_foglamp_remote(self, storage_plugin, remote_user, remote_ip, key_path, remote_foglamp_path):
        """Fixture that kills foglamp, reset database and starts foglamp again on a remote machine
                storage_plugin: Fixture that defines the storage plugin to be used for tests
                remote_user: User of remote machine
                remote_ip: IP of remote machine
                key_path: Path of key file used for authentication to remote machine
                remote_foglamp_path: Path where FogLAMP is cloned and built
            """
        if remote_foglamp_path is None:
            remote_foglamp_path = '/home/{}/foglamp'.format(remote_user)
        subprocess.run(["ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} "
                        "'export FOGLAMP_ROOT={};$FOGLAMP_ROOT/scripts/foglamp kill'".format(
            key_path, remote_user, remote_ip, remote_foglamp_path)], shell=True, check=True)
        storage_plugin_val = "postgres" if storage_plugin == 'postgres' else "sqlite"
        # Check whether storage.json file exist on remote machine or not,
        # if it doesn't exist then raise assertion otherwise update its storage plugin value.
        ssh = subprocess.Popen(["ssh", "-o", "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no",
                                "-i", "{}".format(key_path), "{}@{}".format(remote_user, remote_ip),
                                "cat {}/data/etc/storage.json".format(remote_foglamp_path)], shell=False,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result = ssh.stdout.readlines()
        assert [] != result, "storage.json file not found on the remote machine {}".format(remote_ip)
        data = json.loads(result[0])
        data['plugin']['value'] = storage_plugin_val
        ssh = subprocess.Popen(["ssh", "-o", "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no",
                                "-i", "{}".format(key_path), "{}@{}".format(remote_user, remote_ip),
                                "echo '" + json.dumps(data) + "' > {}/data/etc/storage.json".format(
                                    remote_foglamp_path) ], shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        subprocess.run(["ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} "
                        "'export FOGLAMP_ROOT={};echo \"YES\nYES\" | $FOGLAMP_ROOT/scripts/foglamp reset'".format(
            key_path, remote_user, remote_ip, remote_foglamp_path)], shell=True, check=True)
        # Authentication is optional
        command = ('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} '
                   '"export FOGLAMP_ROOT={} && sed -i \\"s/\'default\': \'mandatory\'/\'default\': \'optional\'/g\\" '
                   '\\$FOGLAMP_ROOT/python/foglamp/services/core/server.py"').format(
            key_path, remote_user, remote_ip, remote_foglamp_path)
        subprocess.run(command, shell=True, check=True)
        subprocess.run(["ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} "
                        "'export FOGLAMP_ROOT={};$FOGLAMP_ROOT/scripts/foglamp start'".format(
            key_path, remote_user, remote_ip, remote_foglamp_path)], shell=True)
        stat = subprocess.run(["ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} "
                               "'export FOGLAMP_ROOT={}; $FOGLAMP_ROOT/scripts/foglamp status'".format(
            key_path, remote_user, remote_ip, remote_foglamp_path)], shell=True, stdout=subprocess.PIPE)
        assert "FogLAMP not running." not in stat.stdout.decode("utf-8")

    @pytest.fixture
    def start_south_north_remote(self, reset_and_start_foglamp_remote, use_pip_cache, remote_user,
                                 key_path, remote_foglamp_path, remote_ip, south_branch,
                                 start_north_pi_server_c_web_api, pi_host, pi_port,
                                 clear_pi_system_through_pi_web_api, pi_admin, pi_passwd, pi_db):
        """Fixture that starts south and north plugins on remote machine
                reset_and_start_foglamp_remote: Fixture that kills foglamp, reset database and starts foglamp again on a remote machine
                use_pip_cache: flag to tell whether to use python's pip cache for python dependencies
                remote_user: User of remote machine
                remote_foglamp_path: Path where FogLAMP is cloned and built
                remote_ip: IP of remote machine
                south_branch: branch of foglamp south plugin
                start_north_pi_server_c_web_api: fixture that configures and starts pi plugin
                pi_host: Host IP of PI machine
                pi_token: Token of connector relay of PI
            """

        # No need to give asset hierarchy in case of connector relay.
        # There are two data points here. 1. ivalue 2. No data point (Asset name is will be used).
        dp_list = ['ivalue', '']
        asset_dict = {}
        asset_dict[remote_asset_name] = dp_list
        # For connector relay we should not delete PI Point because
        # when the PI point is created again (after deletion) the compressing attribute for it
        # is always true. That means all the data is not stored in PI data archive.
        # We lose a large proportion of the data because of compressing attribute.
        # This is problematic for the fixture that verifies the data stored in PI.
        # clear_pi_system_through_pi_web_api(pi_host, pi_admin, pi_passwd, pi_db,
        #                                    [], asset_dict)

        if remote_foglamp_path is None:
            remote_foglamp_path = '/home/{}/foglamp'.format(remote_user)
        foglamp_url = "{}:8081".format(remote_ip)
        south_plugin = "http"
        south_service = "http_south"

        # Install http_south python plugin on remote machine
        try:
            subprocess.run([
                "scp -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} $FOGLAMP_ROOT/tests/system/python/scripts/install_python_plugin {}@{}:/tmp/".format(
                    key_path, remote_user, remote_ip)], shell=True, check=True)
            subprocess.run(["ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {} {}@{} 'export FOGLAMP_ROOT={}; /tmp/install_python_plugin {} south {} {}'".format(
                    key_path, remote_user, remote_ip, remote_foglamp_path, south_branch, south_plugin, use_pip_cache)],
                shell=True, check=True)

        except subprocess.CalledProcessError:
            assert False, "{} plugin installation failed".format(south_plugin)
        conn = http.client.HTTPConnection(foglamp_url)

        # Configure http_south python plugin on remote machine
        data = {"name": "{}".format(south_service), "type": "South", "plugin": "{}".format(south_service),
                "enabled": "true", "config": {"assetNamePrefix": {"value": ""}}}
        conn.request("POST", '/foglamp/service', json.dumps(data))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        retval = json.loads(r)
        assert south_service == retval["name"]

        # Configure pi north plugin on remote machine
        start_north_pi_server_c_web_api(foglamp_url, pi_host, pi_port, pi_db=pi_db, pi_user=pi_admin, pi_pwd=pi_passwd, taskname="NorthReadingsToPI")

        yield self.start_south_north_remote

    def configure_and_start_north_http(self, north_branch, foglamp_url, remote_ip, task_name="NorthReadingsToHTTP"):
        """ Configure and Start north http task """

        try:
            subprocess.run(["$FOGLAMP_ROOT/tests/system/python/scripts/install_c_plugin {} north {}"
                           .format(north_branch, "http-c")], shell=True, check=True)
        except subprocess.CalledProcessError:
            assert False, "http north plugin installation failed"

        conn = http.client.HTTPConnection(foglamp_url)
        data = {"name": task_name,
                "plugin": "{}".format("httpc"),
                "type": "north",
                "schedule_type": 3,
                "schedule_day": 0,
                "schedule_time": 0,
                "schedule_repeat": 30,
                "schedule_enabled": "false",
                "config": {"URL": {"value": "http://{}:6683/sensor-reading".format(remote_ip)}}
                }

        conn.request("POST", '/foglamp/scheduled/task', json.dumps(data))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        val = json.loads(r)
        assert 2 == len(val)
        assert task_name == val['name']

    @pytest.fixture
    def start_south_north_local(self, reset_and_start_foglamp, add_south, enable_schedule, remove_directories,
                                remove_data_file, south_branch, north_branch, foglamp_url, remote_ip,
                                add_filter, filter_branch):
        """ This fixture clone a south and north repo and starts both south and north instance

            reset_and_start_foglamp: Fixture that resets and starts foglamp, no explicit invocation, called at start
            add_south: Fixture that adds a south service with given configuration with enabled or disabled mode
            enable_schedule: Fixture used to enable a schedule
            remove_directories: Fixture that remove directories created during the tests
            remove_data_file: Fixture that remove data file created during the tests
            south_branch: south branch to pull
            north_branch: north branch to pull
            foglamp_url: FogLAMP instance url for local setup (Instance 1)
            remote_ip: IP of remote machine which will receive data from local instance
            add_filter: Fixture that add and configures a filter
            filter_branch: filter branch to pull
        """
        # Add playback plugin
        south_config_playbk = {"assetName": {"value": "{}".format("fogpair_playback")},
                               "csvFilename": {"value": "{}".format(CSV_NAME)},
                               "ingestMode": {"value": "batch"}}

        # Define the CSV data and create expected lists to be verified later
        csv_file_path = os.path.join(os.path.expandvars('${FOGLAMP_ROOT}'), 'data/{}'.format(CSV_NAME))
        with open(csv_file_path, 'w') as f:
            f.write(CSV_HEADERS)
            for _items in CSV_DATA.split(","):
                f.write("\n{}".format(_items))

        south_plugin_playbk = "playback"
        add_south(south_plugin_playbk, south_branch, foglamp_url, service_name="fogpair_playbk",
                  config=south_config_playbk, start_service=False)

        # Add expression plugin
        south_plugin_expression = "Expression"
        south_config_expr = {"expression": {"value": "cos(x)"}, "minimumX": {"value": "45"},
                             "maximumX": {"value": "45"}, "stepX": {"value": "0"}}

        add_south(south_plugin_expression, south_branch, foglamp_url, service_name="fogpair_expr",
                  config=south_config_expr, plugin_lang="C", start_service=False)

        # Add sinusoid plugin
        south_plugin_sinusoid = "sinusoid"

        add_south(south_plugin_sinusoid, south_branch, foglamp_url, service_name="fogpair_sine", start_service=False)

        self.configure_and_start_north_http(north_branch, foglamp_url, remote_ip)

        # Add asset filter
        # I/P All assets > O/P only fogpair_playback asset
        filter_cfg_asset = {"config": {"rules": [{"asset_name": "fogpair_playback", "action": "include"}],
                                       "defaultAction": "exclude"}, "enable": "true"}
        add_filter("asset", filter_branch, "fasset", filter_cfg_asset, foglamp_url, "NorthReadingsToHTTP")

        # Enable all south and north schedules
        enable_schedule(foglamp_url, "fogpair_playbk")
        enable_schedule(foglamp_url, "fogpair_expr")
        enable_schedule(foglamp_url, "fogpair_sine")
        enable_schedule(foglamp_url, "NorthReadingsToHTTP")

        yield self.start_south_north_local

        # Cleanup
        remove_directories("/tmp/foglamp-south-{}".format(south_plugin_playbk))
        remove_directories("/tmp/foglamp-south-{}".format(south_plugin_expression.lower()))
        remove_directories("/tmp/foglamp-south-{}".format(south_plugin_sinusoid))
        remove_directories("/tmp/foglamp-north-{}".format("http"))
        remove_data_file(csv_file_path)

    def _verify_egress(self, read_data_from_pi_asset_server, pi_host, pi_admin, pi_passwd, pi_db, wait_time, retries,
                       expected_read_values):
        """
        Verify that data is received in pi db by making calls to PI web api
            read_data_from_pi_asset_server: Fixture that reads data drom pi
            pi_host: pi host
            pi_admin: pi machine username
            pi_passwd: pi machine password
            pi_db: pi database
            wait_time: wait before making a next call to pi web api
            retries: number of tries to make
            expected_read_values: expected readings value
        """
        retry_count = 0
        data_from_pi = None
        while (data_from_pi is None or data_from_pi == []) and retry_count < retries:
            data_from_pi = read_data_from_pi_asset_server(pi_host, pi_admin, pi_passwd, pi_db, "fogpair_playback", [CSV_HEADERS])
            retry_count += 1
            time.sleep(wait_time * 2)

        if data_from_pi is None or retry_count == retries:
            assert False, "Failed to read data from PI"

        assert Counter(data_from_pi[CSV_HEADERS][-len(expected_read_values):]) == Counter(expected_read_values)

    def test_end_to_end(self, start_south_north_remote, start_south_north_local,
                        read_data_from_pi_asset_server, retries, pi_host, pi_admin, pi_passwd, pi_db,
                        foglamp_url, remote_ip, wait_time, skip_verify_north_interface):
        """ Test that data is inserted in FogLAMP (local instance) using playback south plugin,
            sinusoid south plugin and expression south plugin and sent to http north (filter only playback data),
            FogLAMP (remote instance) receive this data via http south and send to PI
            start_south_north_remote: Fixture that starts FogLAMP with http south service and pi north instance
            start_south_north_local: Fixture that starts FogLAMP with south services and north instance with asset filter
            read_data_from_pi_asset_server: Fixture that reads data from PI web api
            retries: number to retries to make to fetch data from pi
            pi_host: PI host IP
            pi_admin: PI Machine user
            pi_passwd: PI Machine user
            pi_db: PI database
            foglamp_url: Local FogLAMP URL
            remote_ip: IP address where 2 FogLAMP is running (Remote)
            wait_time: time to wait in sec before making assertions
            skip_verify_north_interface: Flag for assertion of data from Pi web API
            Assertions:
                on endpoint GET /foglamp/asset
                on endpoint GET /foglamp/asset/<asset_name> with applied data processing filter value
                data received from PI is same as data sent"""

        # Wait for data to be sent to FogLAMP instance 2 and then to PI
        time.sleep(wait_time * 3)

        # FogLAMP Instance 1 (Local) verification
        expected_asset_list = ["Expression", "fogpair_playback", "sinusoid"]
        actual_asset_list = self.get_asset_list(foglamp_url)
        assert set(expected_asset_list) == set(actual_asset_list)

        ping_response = self.get_ping_status(foglamp_url)
        assert 4 <= ping_response["dataRead"]
        assert 4 == ping_response["dataSent"]

        actual_stats_map = self.get_statistics_map(foglamp_url)
        assert 1 < actual_stats_map['EXPRESSION']
        assert 1 < actual_stats_map['SINUSOID']
        assert 4 == actual_stats_map['FOGPAIR_PLAYBACK']
        assert 4 == actual_stats_map['NorthReadingsToHTTP']
        assert 6 <= actual_stats_map['READINGS']
        assert 4 == actual_stats_map['Readings Sent']

        # FogLAMP Instance 2 (Remote) verification
        foglamp_url_remote = "{}:8081".format(remote_ip)
        conn_remote = http.client.HTTPConnection(foglamp_url_remote)

        expected_list = ["fogpair_playback"]
        actual_asset_list = self.get_asset_list(foglamp_url_remote)
        assert set(expected_list) == set(actual_asset_list)

        remote_ping_response = self.get_ping_status(foglamp_url_remote)
        assert 4 == remote_ping_response["dataRead"]

        actual_stats_map = self.get_statistics_map(foglamp_url_remote)
        assert 'EXPRESSION' not in actual_stats_map.keys()
        assert 'SINUSOID' not in actual_stats_map.keys()
        assert 4 == actual_stats_map['FOGPAIR_PLAYBACK']
        assert 4 == actual_stats_map['READINGS']

        if not skip_verify_north_interface:
            assert 4 == remote_ping_response["dataSent"]
            assert 4 == actual_stats_map['NorthReadingsToPI']
            assert 4 == actual_stats_map['Readings Sent']

        conn_remote.request("GET", '/foglamp/asset/{}'.format("fogpair_playback"))
        r = conn_remote.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        tmp_list = CSV_DATA.split(',')
        tmp_list.reverse()
        expected_read_values = [int(x) for x in tmp_list]
        assert len(expected_read_values) == len(jdoc)

        actual_read_values = []
        for itm in jdoc:
            actual_read_values.append(itm['reading'][CSV_HEADERS])
        assert expected_read_values == actual_read_values

        if not skip_verify_north_interface:
            self._verify_egress(read_data_from_pi_asset_server, pi_host, pi_admin, pi_passwd, pi_db, wait_time, retries,
                                expected_read_values)
